package main

import (
	"bufio"
	"bytes"
	"fmt"
	"io"
	"regexp"
	"strconv"
)

// descHeader is the leading type annotation of CollecTor-format files.
// https://collector.torproject.org/formats.html: "Each descriptor provided here
// contains an @type annotation using the format @type $descriptortype
// $major.$minor".
type descHeader struct {
	Type         string
	Major, Minor uint
}

// descReader is modeled after tar.Reader. Call Next to advance to the next
// descriptor. The Read function reads within the current descriptor.
type descReader struct {
	r   *bufio.Reader
	buf []byte
}

func newDescReader(r io.Reader) *descReader {
	return &descReader{bufio.NewReader(r), nil}
}

func (d *descReader) Read(p []byte) (int, error) {
	var err error

	if len(d.buf) == 0 {
		// Ignore error from Peek; we don't want an EOF from a short
		// final line to be reported as an EOF here. The error will be
		// reported after the call to ReadBytes below.
		l, _ := d.r.Peek(6)
		if bytes.HasPrefix(l, []byte("@type ")) {
			return 0, io.EOF
		}
		d.buf, err = d.r.ReadBytes('\n')
	}

	n := copy(p, d.buf)
	d.buf = d.buf[n:]
	return n, err
}

var typeRE = regexp.MustCompile("^@type ([\\w-]+) ([0-9]+)\\.([0-9]+)\n$")

func parseUint(s string) (uint, error) {
	n, err := strconv.ParseUint(s, 10, 0)
	return uint(n), err
}

func parseTypeAnnotation(l []byte) (descType string, major, minor uint, err error) {
	matches := typeRE.FindSubmatch(l)
	if matches == nil {
		err = fmt.Errorf("bad @type annotation %q", l)
		return
	}
	descType = string(matches[1])
	major, err = parseUint(string(matches[2]))
	if err != nil {
		return
	}
	minor, err = parseUint(string(matches[3]))
	if err != nil {
		return
	}
	return
}

func (d *descReader) Next() (*descHeader, error) {
	var h descHeader
	var l []byte
	var err error

	// Read lines until the next type annotation.
	d.buf = nil
	for {
		l, err = d.r.ReadBytes('\n')
		if err != nil {
			return nil, err
		}
		if bytes.HasPrefix(l, []byte("@type ")) {
			break
		}
	}

	h.Type, h.Major, h.Minor, err = parseTypeAnnotation(l)
	return &h, nil
}
