package main

import (
	"bufio"
	"encoding/base64"
	"encoding/csv"
	"encoding/hex"
	"flag"
	"fmt"
	"io"
	"log"
	"os"
	"strings"
	"time"
)

type routerStatus struct {
	Fingerprint  string
	Flags        map[string]struct{}
	Bandwidth    uint
	RouterDigest string
}

func (rs *routerStatus) RunningAndValid() bool {
	for _, flag := range []string{"Running", "Valid"} {
		_, ok := rs.Flags[flag]
		if !ok {
			return false
		}
	}
	return true
}

func parseBase64Digest(b64 string) (string, error) {
	b, err := base64.StdEncoding.DecodeString(b64 + "==="[:3-(len(b64)+3)%4])
	return string(b), err
}

func parseHexDigest(h string) (string, error) {
	b, err := hex.DecodeString(h)
	return string(b), err
}

func maybeOutputRouterStatus(published time.Time, rs *routerStatus) error {
	if !rs.RunningAndValid() {
		return nil
	}
	if published.IsZero() {
		return fmt.Errorf("missing %q line before %q line", "published", "r")
	}
	return bridgeNetworkStatusCSV.Write([]string{
		published.Format("2006-01-02"),
		published.Format("15:04:05"),
		fmt.Sprintf("%X", rs.Fingerprint),
		fmt.Sprintf("%d", rs.Bandwidth),
		fmt.Sprintf("%X", rs.RouterDigest),
	})
}

func parseBridgeNetworkStatus(r io.Reader) error {
	var published time.Time
	var err error
	var rs *routerStatus

	s := bufio.NewScanner(r)
	for s.Scan() {
		l := s.Text()
		i := strings.IndexByte(l, ' ')
		if i == -1 {
			i = len(l)
		}
		keyword := l[:i]

		switch keyword {
		case "published", "r", "s", "w":
		default:
			// Ignore all other keywords.
			continue
		}

		fields := strings.Fields(l[i:])
		switch keyword {
		case "published":
			if !published.IsZero() {
				return fmt.Errorf("duplicate %q line: %q", keyword, l)
			}
			published, err = time.Parse("2006-01-02 15:04:05", fields[0]+" "+fields[1])
			if err != nil {
				return err
			}
		case "r":
			if rs != nil {
				err = maybeOutputRouterStatus(published, rs)
				if err != nil {
					return err
				}
			}
			rs = new(routerStatus)
			rs.Flags = make(map[string]struct{})
			rs.Fingerprint, err = parseBase64Digest(fields[1])
			if err != nil {
				return err
			}
			rs.RouterDigest, err = parseBase64Digest(fields[2])
			if err != nil {
				return err
			}
		case "s":
			if rs == nil {
				return fmt.Errorf("%q line before %q line", keyword, "r")
			}
			for _, f := range fields {
				rs.Flags[f] = struct{}{}
			}
		case "w":
			if rs == nil {
				return fmt.Errorf("%q line before %q line", keyword, "r")
			}
			for _, f := range fields {
				parts := strings.SplitN(f, "=", 2)
				if len(parts) != 2 {
					return fmt.Errorf("weird %q entry %q", keyword, f)
				}
				if parts[0] == "Bandwidth" {
					rs.Bandwidth, err = parseUint(parts[1])
					if err != nil {
						return fmt.Errorf("bad Bandwidth value: %q", parts[1])
					}
				}
			}
		}
	}
	if s.Err() != nil {
		return s.Err()
	}

	if rs != nil {
		err = maybeOutputRouterStatus(published, rs)
		if err != nil {
			return err
		}
	}

	return nil
}

func parseBridgeServerDescriptor(r io.Reader) error {
	var err error
	var routerDigest, extraInfoDigest string

	s := bufio.NewScanner(r)
	for s.Scan() {
		l := s.Text()
		i := strings.IndexByte(l, ' ')
		if i == -1 {
			i = len(l)
		}
		keyword := l[:i]

		switch keyword {
		case "router-digest", "extra-info-digest":
		default:
			// Ignore all other keywords.
			continue
		}

		fields := strings.Fields(l[i:])
		switch keyword {
		case "router-digest":
			if routerDigest != "" {
				return fmt.Errorf("duplicate %q line: %q", keyword, l)
			}
			routerDigest, err = parseHexDigest(fields[0])
			if err != nil {
				return err
			}
		case "extra-info-digest":
			if extraInfoDigest != "" {
				return fmt.Errorf("duplicate %q line: %q", keyword, l)
			}
			extraInfoDigest, err = parseHexDigest(fields[0])
			if err != nil {
				return err
			}
		}
	}
	if s.Err() != nil {
		return s.Err()
	}

	if routerDigest == "" {
		return fmt.Errorf("no %q line", "router-digest")
	}

	if extraInfoDigest == "" {
		return nil
	}
	return bridgeServerDescriptorCSV.Write([]string{
		fmt.Sprintf("%X", routerDigest),
		fmt.Sprintf("%X", extraInfoDigest),
	})
}

// Append a string, filtering duplicates.
func appendStringUnique(l []string, s string) []string {
	for _, t := range l {
		if s == t {
			return l
		}
	}
	return append(l, s)
}

func parseBridgeExtraInfo(r io.Reader) error {
	var err error
	var extraInfoDigest string
	var transports []string

	s := bufio.NewScanner(r)
	for s.Scan() {
		l := s.Text()
		i := strings.IndexByte(l, ' ')
		if i == -1 {
			i = len(l)
		}
		keyword := l[:i]

		switch keyword {
		case "transport", "router-digest":
		default:
			// Ignore all other keywords.
			continue
		}

		fields := strings.Fields(l[i:])
		switch keyword {
		case "transport":
			transports = appendStringUnique(transports, fields[0])
		case "router-digest":
			if extraInfoDigest != "" {
				return fmt.Errorf("duplicate %q line: %q", keyword, l)
			}
			extraInfoDigest, err = parseHexDigest(fields[0])
			if err != nil {
				return err
			}
		}
	}
	if s.Err() != nil {
		return s.Err()
	}

	if extraInfoDigest == "" {
		return fmt.Errorf("no %q line in bridge-extra-info document", "router-digest")
	}

	for _, transport := range transports {
		err = bridgeExtraInfoCSV.Write([]string{
			fmt.Sprintf("%X", extraInfoDigest),
			transport,
		})
		if err != nil {
			return err
		}
	}

	return nil
}

func process(f io.Reader) error {
	d := newDescReader(f)
	for {
		h, err := d.Next()
		if err != nil {
			if err == io.EOF {
				break
			}
			return err
		}
		// fmt.Printf("@type %s %d.%d\n", h.Type, h.Major, h.Minor)
		switch h.Type {
		case "bridge-network-status":
			err = parseBridgeNetworkStatus(d)
		case "bridge-server-descriptor":
			err = parseBridgeServerDescriptor(d)
		case "bridge-extra-info":
			err = parseBridgeExtraInfo(d)
		}
		if err != nil {
			return err
		}
	}

	return nil
}

func newTarXZReader(r io.Reader) (io.Reader, error) {
	xzr, err := newXZReader(r)
	if err != nil {
		return nil, err
	}
	return newTarConcatReader(xzr)
}

func processTarXZFile(filename string) error {
	f, err := os.Open(filename)
	if err != nil {
		return err
	}
	defer f.Close()

	r, err := newTarXZReader(f)
	if err != nil {
		return err
	}

	return process(r)
}

var bridgeNetworkStatusCSV, bridgeServerDescriptorCSV, bridgeExtraInfoCSV *csv.Writer

func main() {
	flag.Parse()

	var f *os.File
	var err error

	f, err = os.Create("bridge-network-status.csv")
	if err != nil {
		log.Fatal(err)
	}
	defer f.Close()
	bridgeNetworkStatusCSV = csv.NewWriter(f)
	defer bridgeNetworkStatusCSV.Flush()
	err = bridgeNetworkStatusCSV.Write([]string{"date", "time", "hashedfp", "bandwidth", "router-digest"})
	if err != nil {
		log.Fatal(err)
	}

	f, err = os.Create("bridge-server-descriptor.csv")
	if err != nil {
		log.Fatal(err)
	}
	defer f.Close()
	bridgeServerDescriptorCSV = csv.NewWriter(f)
	defer bridgeServerDescriptorCSV.Flush()
	err = bridgeServerDescriptorCSV.Write([]string{"router-digest", "extra-info-digest"})
	if err != nil {
		log.Fatal(err)
	}

	f, err = os.Create("bridge-extra-info.csv")
	if err != nil {
		log.Fatal(err)
	}
	defer f.Close()
	bridgeExtraInfoCSV = csv.NewWriter(f)
	defer bridgeExtraInfoCSV.Flush()
	err = bridgeExtraInfoCSV.Write([]string{"extra-info-digest", "transport"})
	if err != nil {
		log.Fatal(err)
	}

	for _, filename := range flag.Args() {
		err := processTarXZFile(filename)
		if err != nil {
			log.Fatal(err)
		}
	}
}
